# Copyright (c) 2025, Tri Dao.
# Shared utilities for TVM-FFI GEMM compilation.

from functools import partial


import cutlass.cute as cute
from cutlass import Int32, Int64, Float32
from cutlass.cute.runtime import make_ptr

from .compile_utils import make_fake_tensor as fake_tensor
from .cute_dsl_utils import torch2cute_dtype_map
from .tile_scheduler import TileSchedulerOptions
from .varlen_utils import VarlenArguments, VarlenNArguments


def div_for_dtype(dtype):
    """16-byte alignment: divisibility in elements = 128 // dtype_width_bits."""
    return 128 // dtype.width


def perm3d_single(t, varlen_m=False):
    """Permute a single 3D tensor from (L, *, *) to (*, *, L), skipping for varlen_m or 2D."""
    return t.permute(1, 2, 0) if t is not None and t.ndim == 3 and not varlen_m else t


def perm3d(A, B, D, C, varlen_m=False, varlen_k=False, varlen_n=False):
    """Permute 3D tensors from (L, *, *) to (*, *, L)."""

    def _perm(t):
        return t.permute(1, 2, 0) if t is not None and t.ndim == 3 else t

    if varlen_m:
        return A, _perm(B), D, C
    elif varlen_k:
        return A, B, _perm(D), _perm(C)
    elif varlen_n:
        return _perm(A), B, D, C
    else:
        return _perm(A), _perm(B), _perm(D), _perm(C)


def get_major(t, dim0, dim1):
    return dim1 if t.stride(1) == 1 else dim0


def get_majors(A_p, B_p, D_p, C_p):
    a_major = get_major(A_p, "m", "k")
    b_major = get_major(B_p, "n", "k")
    d_major = get_major(D_p, "m", "n")
    c_major = get_major(C_p, "m", "n") if C_p is not None else None
    return a_major, b_major, d_major, c_major


def get_dtypes(A, B, D, C):
    a_dtype = torch2cute_dtype_map[A.dtype]
    b_dtype = torch2cute_dtype_map[B.dtype]
    d_dtype = torch2cute_dtype_map[D.dtype]
    c_dtype = torch2cute_dtype_map[C.dtype] if C is not None else None
    return a_dtype, b_dtype, d_dtype, c_dtype


def make_scheduler_args(
    max_active_clusters, max_swizzle_size, tile_count_semaphore, batch_idx_permute=None
):
    return TileSchedulerOptions(
        max_active_clusters=Int32(max_active_clusters),
        raster_order=None,
        max_swizzle_size=max_swizzle_size,
        tile_count_semaphore=(
            tile_count_semaphore.data_ptr() if tile_count_semaphore is not None else None
        ),
        batch_idx_permute=batch_idx_permute,
    )


def make_fake_scheduler_args(has_semaphore, has_batch_idx_permute, l_sym):
    return TileSchedulerOptions(
        max_active_clusters=Int32(1),
        max_swizzle_size=Int32(8),
        tile_count_semaphore=(
            make_ptr(Int32, 0, cute.AddressSpace.gmem, assumed_align=4) if has_semaphore else None
        ),
        batch_idx_permute=(
            fake_tensor(Int32, (l_sym,), leading_dim=0, divisibility=4)
            if has_batch_idx_permute
            else None
        ),
    )


def make_varlen_args(cu_seqlens_m, cu_seqlens_k, A_idx=None, *, cu_seqlens_n=None):
    if cu_seqlens_m is None and cu_seqlens_k is None and cu_seqlens_n is None:
        return None
    if cu_seqlens_n is not None:
        return VarlenNArguments(
            mCuSeqlensM=cu_seqlens_m,
            mCuSeqlensK=cu_seqlens_k,
            mAIdx=A_idx,
            mCuSeqlensN=cu_seqlens_n,
        )
    return VarlenArguments(
        mCuSeqlensM=cu_seqlens_m,
        mCuSeqlensK=cu_seqlens_k,
        mAIdx=A_idx,
    )


def make_fake_varlen_args(varlen_m, varlen_k, gather_A=False, aidx_len=None, *, varlen_n=False):
    if not varlen_m and not varlen_k and not varlen_n:
        return None
    num_seqlens = cute.sym_int()
    cu_seqlens_m = (
        fake_tensor(Int32, (num_seqlens,), leading_dim=0, divisibility=4) if varlen_m else None
    )
    cu_seqlens_k = (
        fake_tensor(Int32, (num_seqlens,), leading_dim=0, divisibility=4) if varlen_k else None
    )
    a_idx = (
        fake_tensor(Int32, (aidx_len,), leading_dim=0, divisibility=4)
        if gather_A is True
        else None
    )
    if varlen_n:
        return VarlenNArguments(
            mCuSeqlensM=cu_seqlens_m,
            mCuSeqlensK=cu_seqlens_k,
            mAIdx=a_idx,
            mCuSeqlensN=fake_tensor(Int32, (num_seqlens,), leading_dim=0, divisibility=4),
        )
    return VarlenArguments(
        mCuSeqlensM=cu_seqlens_m,
        mCuSeqlensK=cu_seqlens_k,
        mAIdx=a_idx,
    )


def make_fake_gemm_tensors(
    a_dtype,
    b_dtype,
    d_dtype,
    c_dtype,
    a_major,
    b_major,
    d_major,
    c_major,
    varlen_m=False,
    varlen_k=False,
    varlen_n=False,
    gather_A=False,
):
    """Create fake tensors for mA, mB, mD, mC with shared sym_ints.
    Pass dtype=None to get None for that tensor (e.g. optional C).
    Returns (mA, mB, mD, mC, m, n, k, l).
    When varlen_m, m is total_m (flattened M of D/C). When varlen_k, k is total_k.
    """
    a_leading = 1 if a_major == "k" else 0
    b_leading = 1 if b_major == "k" else 0
    d_leading = 1 if d_major == "n" else 0
    c_leading = 1 if c_major == "n" else 0
    m, n, k, l = cute.sym_int(), cute.sym_int(), cute.sym_int(), cute.sym_int()
    div_a = div_for_dtype(a_dtype)
    div_b = div_for_dtype(b_dtype)
    div_d = div_for_dtype(d_dtype) if d_dtype is not None else 1
    div_c = div_for_dtype(c_dtype) if c_dtype is not None else 1
    if varlen_m:
        # m is total_m in this case: the flattened M dimension of D/C
        m = cute.sym_int()
        a_m = cute.sym_int() if gather_A else m
        mA = fake_tensor(a_dtype, (a_m, k), leading_dim=a_leading, divisibility=div_a)
        mB = fake_tensor(b_dtype, (n, k, l), leading_dim=b_leading, divisibility=div_b)
        mD = fake_tensor(d_dtype, (m, n), leading_dim=d_leading, divisibility=div_d)
        mC = fake_tensor(c_dtype, (m, n), leading_dim=c_leading, divisibility=div_c)
    elif varlen_k:
        # k is total_k in this case: the flattened K dimension of A/B
        k = cute.sym_int()
        a_k = cute.sym_int() if gather_A else k
        mA = fake_tensor(a_dtype, (m, a_k), leading_dim=a_leading, divisibility=div_a)
        mB = fake_tensor(b_dtype, (n, k), leading_dim=b_leading, divisibility=div_b)
        mD = fake_tensor(d_dtype, (m, n, l), leading_dim=d_leading, divisibility=div_d)
        mC = fake_tensor(c_dtype, (m, n, l), leading_dim=c_leading, divisibility=div_c)
    elif varlen_n:
        n = cute.sym_int()
        mA = fake_tensor(a_dtype, (m, k, l), leading_dim=a_leading, divisibility=div_a)
        mB = fake_tensor(b_dtype, (n, k), leading_dim=b_leading, divisibility=div_b)
        mD = fake_tensor(d_dtype, (m, n), leading_dim=d_leading, divisibility=div_d)
        mC = fake_tensor(c_dtype, (m, n), leading_dim=c_leading, divisibility=div_c)
    else:
        mA = fake_tensor(a_dtype, (m, k, l), leading_dim=a_leading, divisibility=div_a)
        mB = fake_tensor(b_dtype, (n, k, l), leading_dim=b_leading, divisibility=div_b)
        mD = fake_tensor(d_dtype, (m, n, l), leading_dim=d_leading, divisibility=div_d)
        mC = fake_tensor(c_dtype, (m, n, l), leading_dim=c_leading, divisibility=div_c)
    return mA, mB, mD, mC, m, n, k, l


def compile_gemm_kernel(
    GemmCls,
    a_dtype,
    tile_shape_mn,
    cluster_shape_mnk,
    pingpong,
    persistent,
    gather_A,
    is_dynamic_persistent,
    device_capacity,
    mA,
    mB,
    mD,
    mC,
    epi_args,
    scheduler_args,
    varlen_args,
    post_init=None,
    mSFA=None,
    mSFB=None,
    has_trace_ptr=False,
    use_tma_gather=False,
    concat_layout=None,
    num_warps=None,
):
    """Build GemmCls instance, apply SM90 partial, and cute.compile with TVM-FFI."""
    if device_capacity[0] == 8:
        sm8x_kwargs = {"is_persistent": persistent, "num_warps": num_warps}
        sm8x_kwargs["arch"] = device_capacity[0] * 10 + device_capacity[1]
        GemmCls = partial(GemmCls, **sm8x_kwargs)
    elif device_capacity[0] in [9, 12]:
        GemmCls = partial(GemmCls, pingpong=pingpong, is_persistent=persistent)
    elif device_capacity[0] in [10, 11]:
        GemmCls = partial(
            GemmCls,
            use_clc_persistence=is_dynamic_persistent,
            use_tma_gather=use_tma_gather,
        )
    gemm_obj = GemmCls(
        Float32,
        a_dtype,
        tile_shape_mn,
        cluster_shape_mnk,
        gather_A=gather_A,
        concat_layout=concat_layout,
    )
    if post_init:
        post_init(gemm_obj)
    stream = cute.runtime.make_fake_stream(use_tvm_ffi_env_stream=True)
    sf_args = () if device_capacity[0] in (8, 9, 12) else (mSFA, mSFB)
    # Trace pointer: Optional[Int64]. Compile with Int64(0) when tracing is
    # requested, None otherwise. TVM-FFI caches each variant separately.
    trace_ptr = Int64(0) if has_trace_ptr else None
    return cute.compile(
        gemm_obj,
        mA,
        mB,
        mD,
        mC,
        epi_args,
        scheduler_args,
        varlen_args,
        stream,
        *sf_args,
        trace_ptr,
        options="--enable-tvm-ffi",
    )



# mypy: allow-untyped-defs
from __future__ import annotations


__all__ = [
    # Modules
    "errors",
    "ops",
    # Public functions
    "export",
    "is_in_onnx_export",
    # Base error
    "OnnxExporterError",
    "ONNXProgram",
    "ExportableModule",
    "InputObserver",
]

from typing import Any, TYPE_CHECKING

import torch
from torch._C import _onnx as _C_onnx
from torch._C._onnx import (  # Deprecated members that are excluded from __all__
    OperatorExportTypes as OperatorExportTypes,
    TensorProtoDataType as TensorProtoDataType,
    TrainingMode as TrainingMode,
)

from . import errors, ops
from ._internal.exporter._exportable_module import ExportableModule
from ._internal.exporter._input_observer import InputObserver
from ._internal.exporter._onnx_program import ONNXProgram
from ._internal.torchscript_exporter import (  # Deprecated members that are excluded from __all__
    symbolic_helper,
    symbolic_opset10,
    symbolic_opset9,
    utils,
)
from ._internal.torchscript_exporter._type_utils import (
    JitScalarType,  # Deprecated members that are excluded from __all__
)
from ._internal.torchscript_exporter.utils import (  # Deprecated members that are excluded from __all__
    register_custom_op_symbolic,
    select_model_mode_for_export,  # pyrefly: ignore  # deprecated
    unregister_custom_op_symbolic,
)
from .errors import OnnxExporterError


if TYPE_CHECKING:
    import os
    from collections.abc import Callable, Collection, Mapping, Sequence

# Set namespace for exposed private names
ONNXProgram.__module__ = "torch.onnx"
ExportableModule.__module__ = "torch.onnx"
OnnxExporterError.__module__ = "torch.onnx"
InputObserver.__module__ = "torch.onnx"

# TODO(justinchuby): Remove these two properties
producer_name = "pytorch"
producer_version = _C_onnx.PRODUCER_VERSION


def export(
    model: torch.nn.Module
    | torch.export.ExportedProgram
    | torch.jit.ScriptModule
    | torch.jit.ScriptFunction,
    args: tuple[Any, ...] = (),
    f: str | os.PathLike | None = None,
    *,
    kwargs: dict[str, Any] | None = None,
    verbose: bool | None = None,
    input_names: Sequence[str] | None = None,
    output_names: Sequence[str] | None = None,
    opset_version: int | None = None,
    dynamo: bool = True,
    # Dynamo only options
    external_data: bool = True,
    dynamic_shapes: dict[str, Any] | tuple[Any, ...] | list[Any] | None = None,
    custom_translation_table: dict[Callable, Callable] | None = None,
    report: bool = False,
    optimize: bool = True,
    verify: bool = False,
    profile: bool = False,
    dump_exported_program: bool = False,
    artifacts_dir: str | os.PathLike = ".",
    # BC options
    export_params: bool = True,
    keep_initializers_as_inputs: bool = False,
    dynamic_axes: Mapping[str, Mapping[int, str]]
    | Mapping[str, Sequence[int]]
    | None = None,
    # Deprecated options
    training: _C_onnx.TrainingMode = _C_onnx.TrainingMode.EVAL,
    operator_export_type: _C_onnx.OperatorExportTypes = _C_onnx.OperatorExportTypes.ONNX,
    do_constant_folding: bool = True,
    custom_opsets: Mapping[str, int] | None = None,
    export_modules_as_functions: bool | Collection[type[torch.nn.Module]] = False,
    autograd_inlining: bool = True,
) -> ONNXProgram | None:
    r"""Exports a model into ONNX format.

    Setting ``dynamo=True`` enables the new ONNX export logic
    which is based on :class:`torch.export.ExportedProgram` and a more modern
    set of translation logic. This is the recommended and default way to export models
    to ONNX.

    When ``dynamo=True``:

    The exporter tries the following strategies to get an ExportedProgram for conversion to ONNX.

    #. If the model is already an ExportedProgram, it will be used as-is.
    #. Use :func:`torch.export.export` and set ``strict=False``.
    #. Use :func:`torch.export.export` and set ``strict=True``.

    Args:
        model: The model to be exported.
        args: Example positional inputs. Any non-Tensor arguments will be hard-coded into the
            exported model; any Tensor arguments will become inputs of the exported model,
            in the order they occur in the tuple.
        f: Path to the output ONNX model file. E.g. "model.onnx". This argument is kept for
            backward compatibility. It is recommended to leave unspecified (None)
            and use the returned :class:`torch.onnx.ONNXProgram` to serialize the model
            to a file instead.
        kwargs: Optional example keyword inputs.
        verbose: Whether to enable verbose logging.
        input_names: names to assign to the input nodes of the graph, in order.
        output_names: names to assign to the output nodes of the graph, in order.
        opset_version: The version of the
            `default (ai.onnx) opset <https://github.com/onnx/onnx/blob/master/docs/Operators.md>`_
            to target. You should set ``opset_version`` according to the supported opset versions
            of the runtime backend or compiler you want to run the exported model with.
            Leave as default (``None``) to use the recommended version, or refer to
            the ONNX operators documentation for more information.
        dynamo: Whether to export the model with ``torch.export`` ExportedProgram instead of TorchScript.
        external_data: Whether to save the model weights as an external data file.
            This is required for models with large weights that exceed the ONNX file size limit (2GB).
            When False, the weights are saved in the ONNX file with the model architecture.
        dynamic_shapes: A dictionary or a tuple of dynamic shapes for the model inputs. Refer to
            :func:`torch.export.export` for more details. This is only used (and preferred) when dynamo is True.
            Note that dynamic_shapes is designed to be used when the model is exported with dynamo=True, while
            dynamic_axes is used when dynamo=False.
        custom_translation_table: A dictionary of custom decompositions for operators in the model.
            The dictionary should have the callable target in the fx Node as the key (e.g. ``torch.ops.aten.stft.default``),
            and the value should be a function that builds that graph using ONNX Script. This option
            is only valid when dynamo is True.
        report: Whether to generate a markdown report for the export process. This option
            is only valid when dynamo is True.
        optimize: Whether to optimize the exported model. This option
            is only valid when dynamo is True. Default is True.
        verify: Whether to verify the exported model using ONNX Runtime. This option
            is only valid when dynamo is True.
        profile: Whether to profile the export process. This option
            is only valid when dynamo is True.
        dump_exported_program: Whether to dump the :class:`torch.export.ExportedProgram` to a file.
            This is useful for debugging the exporter. This option is only valid when dynamo is True.
        artifacts_dir: The directory to save the debugging artifacts like the report and the serialized
            exported program. This option is only valid when dynamo is True.
        export_params: **When ``f`` is specified**: If false, parameters (weights) will not be exported.

            You can also leave it unspecified and use the returned :class:`torch.onnx.ONNXProgram`
            to control how initializers are treated when serializing the model.
        keep_initializers_as_inputs: **When ``f`` is specified**: If True, all the
            initializers (typically corresponding to model weights) in the
            exported graph will also be added as inputs to the graph. If False,
            then initializers are not added as inputs to the graph, and only
            the user inputs are added as inputs.

            Set this to True if you intend to supply model weights at runtime.
            Set it to False if the weights are static to allow for better optimizations
            (e.g. constant folding) by backends/runtimes.

            You can also leave it unspecified and use the returned :class:`torch.onnx.ONNXProgram`
            to control how initializers are treated when serializing the model.
        dynamic_axes:
            Deprecated: Prefer specifying ``dynamic_shapes`` when ``dynamo=True``.

            By default the exported model will have the shapes of all input and output tensors
            set to exactly match those given in ``args``. To specify axes of tensors as
            dynamic (i.e. known only at run-time), set ``dynamic_axes`` to a dict with schema:

            * KEY (str): an input or output name. Each name must also be provided in ``input_names`` or
                ``output_names``.
            * VALUE (dict or list): If a dict, keys are axis indices and values are axis names. If a
                list, each element is an axis index.

            For example::

                class SumModule(torch.nn.Module):
                    def forward(self, x):
                        return torch.sum(x, dim=1)


                torch.onnx.export(
                    SumModule(),
                    (torch.ones(2, 2),),
                    "onnx.pb",
                    input_names=["x"],
                    output_names=["sum"],
                )

            Produces::

                input {
                  name: "x"
                  ...
                      shape {
                        dim {
                          dim_value: 2  # axis 0
                        }
                        dim {
                          dim_value: 2  # axis 1
                ...
                output {
                  name: "sum"
                  ...
                      shape {
                        dim {
                          dim_value: 2  # axis 0
                ...

            While::

                torch.onnx.export(
                    SumModule(),
                    (torch.ones(2, 2),),
                    "onnx.pb",
                    input_names=["x"],
                    output_names=["sum"],
                    dynamic_axes={
                        # dict value: manually named axes
                        "x": {0: "my_custom_axis_name"},
                        # list value: automatic names
                        "sum": [0],
                    },
                )

            Produces::

                input {
                  name: "x"
                  ...
                      shape {
                        dim {
                          dim_param: "my_custom_axis_name"  # axis 0
                        }
                        dim {
                          dim_value: 2  # axis 1
                ...
                output {
                  name: "sum"
                  ...
                      shape {
                        dim {
                          dim_param: "sum_dynamic_axes_1"  # axis 0
                ...

        training: Deprecated option. Instead, set the training mode of the model before exporting.
        operator_export_type: Deprecated option. Only ONNX is supported.
        do_constant_folding: Deprecated option.
        custom_opsets: Deprecated option.
        export_modules_as_functions: Deprecated option.
        autograd_inlining: Deprecated option.

    Returns:
        :class:`torch.onnx.ONNXProgram` if dynamo is True, otherwise None.

    .. versionchanged:: 2.6
        ``training`` is now deprecated. Instead, set the training mode of the model before exporting.
        ``operator_export_type`` is now deprecated. Only ONNX is supported.
        ``do_constant_folding`` is now deprecated. It is always enabled.
        ``export_modules_as_functions`` is now deprecated.
        ``autograd_inlining`` is now deprecated.
    .. versionchanged:: 2.7
        ``optimize`` is now True by default.
    .. versionchanged:: 2.9
        ``dynamo`` is now True by default.
    .. versionchanged:: 2.11
        ``fallback`` option has been removed.
    """
    if dynamo is True or isinstance(
        model, (torch.export.ExportedProgram, ExportableModule)
    ):
        from torch.onnx._internal.exporter import _compat

        if isinstance(args, torch.Tensor):
            args = (args,)

        return _compat.export_compat(
            model,
            args,
            f,
            kwargs=kwargs,
            export_params=export_params,
            verbose=verbose,
            input_names=input_names,
            output_names=output_names,
            opset_version=opset_version,
            custom_translation_table=custom_translation_table,
            dynamic_axes=dynamic_axes,
            keep_initializers_as_inputs=keep_initializers_as_inputs,
            external_data=external_data,
            dynamic_shapes=dynamic_shapes,
            report=report,
            optimize=optimize,
            verify=verify,
            profile=profile,
            dump_exported_program=dump_exported_program,
            artifacts_dir=artifacts_dir,
        )
    else:
        import warnings

        from ._internal.torchscript_exporter.utils import export

        warnings.warn(
            "You are using the legacy TorchScript-based ONNX export. Starting in PyTorch 2.9, "
            "the new torch.export-based ONNX exporter has become the default. "
            "Learn more about the new export logic: https://docs.pytorch.org/docs/stable/onnx_export.html. "
            "For exporting control flow: "
            "https://pytorch.org/tutorials/beginner/onnx/export_control_flow_model_to_onnx_tutorial.html",
            category=DeprecationWarning,
            stacklevel=2,
        )

        if dynamic_shapes:
            raise ValueError(
                "The exporter only supports dynamic shapes "
                "through parameter dynamic_axes when dynamo=False."
            )

        export(
            model,
            args,
            f,  # type: ignore[arg-type]
            kwargs=kwargs,
            export_params=export_params,
            verbose=verbose is True,
            input_names=input_names,
            output_names=output_names,
            opset_version=opset_version,
            dynamic_axes=dynamic_axes,
            keep_initializers_as_inputs=keep_initializers_as_inputs,
            training=training,
            operator_export_type=operator_export_type,
            do_constant_folding=do_constant_folding,
            custom_opsets=custom_opsets,
            export_modules_as_functions=export_modules_as_functions,
            autograd_inlining=autograd_inlining,
        )
        return None


def is_in_onnx_export() -> bool:
    """Returns whether it is in the middle of ONNX export."""
    from torch.onnx._internal.exporter import _flags
    from torch.onnx._internal.torchscript_exporter._globals import GLOBALS

    return GLOBALS.in_onnx_export or _flags._is_onnx_exporting



"""torchgen

This module contains codegeneration utilities for PyTorch. It is used to
build PyTorch from source, but may also be used for out-of-tree projects
that extend PyTorch.

Note well that we provide no BC guarantees for torchgen. If you're interested
in using torchgen and want the PyTorch team to be aware, please reach out
on GitHub.
"""



# mypy: allow-untyped-defs
from __future__ import annotations

from typing import Any, TYPE_CHECKING

from ..scheduler import (
    BaseSchedulerNode,
    BaseScheduling,
    FusedSchedulerNode,
    Scheduler,
    SchedulerNode,
)
from .cutedsl.cutedsl_scheduling import CuteDSLScheduling
from .cutlass.scheduling import CUTLASSScheduling
from .nv_universal_gemm.nv_universal_gemm_scheduling import NVUniversalGemmScheduling
from .rocm.rocm_cpp_scheduling import ROCmCPPScheduling
from .triton import TritonScheduling


if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TypeAlias

    from sympy import Expr

    import torch
    from torch.utils._ordered_set import OrderedSet

    from .common import BackendFeature

    _IntLike: TypeAlias = int | Expr


class CUDACombinedScheduling(BaseScheduling):
    """
    Scheduler for CUDA Kernels, which delegates calls as appropriate
    to the CUDA-C++ and Triton Schedulers, which both work for CUDA devices
    and use a unified-wrapper for codegen.

    If Scheduling code needs to be specialized for the case of mixed Triton / CUDA C++ code,
    this would also be the place to do it.
    """

    def __init__(self, scheduler: Scheduler | None) -> None:
        super().__init__(scheduler)
        self._triton_scheduling = TritonScheduling(scheduler)
        self._cutlass_scheduling = CUTLASSScheduling(scheduler)
        self._rocm_cpp_scheduling = ROCmCPPScheduling(scheduler)
        self._cutedsl_scheduling = CuteDSLScheduling(scheduler)
        self._nv_universal_gemm_scheduling = NVUniversalGemmScheduling(scheduler)

    def get_backend_features(self, device: torch.device) -> OrderedSet[BackendFeature]:
        return self._triton_scheduling.get_backend_features(device)

    def choose_node_backend(self, node: BaseSchedulerNode) -> BaseScheduling:
        if self._cutlass_scheduling.is_cutlass_template(node):
            return self._cutlass_scheduling
        if self._rocm_cpp_scheduling.is_rocm_cpp_template(node):
            return self._rocm_cpp_scheduling
        if self._cutedsl_scheduling.is_cutedsl_template(node):
            return self._cutedsl_scheduling
        if self._nv_universal_gemm_scheduling.is_nv_universal_gemm_template(node):
            return self._nv_universal_gemm_scheduling
        return self._triton_scheduling

    def can_fuse_vertical(
        self, node1: BaseSchedulerNode, node2: BaseSchedulerNode
    ) -> bool:
        if self._cutlass_scheduling.can_fuse_vertical(node1, node2):
            return True
        elif self._cutlass_scheduling.is_cutlass_template(
            node1
        ) or self._cutlass_scheduling.is_cutlass_template(node2):
            return False
        # CuteDSL doesn't support vertical fusion currently
        elif self._cutedsl_scheduling.is_cutedsl_template(
            node1
        ) or self._cutedsl_scheduling.is_cutedsl_template(node2):
            return False
        # NVIDIA Universal GEMM doesn't support vertical fusion currently
        elif self._nv_universal_gemm_scheduling.is_nv_universal_gemm_template(
            node1
        ) or self._nv_universal_gemm_scheduling.is_nv_universal_gemm_template(node2):
            return False
        return self._triton_scheduling.can_fuse_vertical(node1, node2)

    def can_fuse_horizontal(
        self, node1: BaseSchedulerNode, node2: BaseSchedulerNode
    ) -> bool:
        for node in (node1, node2):
            if self._cutlass_scheduling.is_cutlass_template(node):
                return self._cutlass_scheduling.can_fuse_horizontal(
                    node1, node2
                )  # always False at the moment
            if self._cutedsl_scheduling.is_cutedsl_template(node):
                return self._cutedsl_scheduling.can_fuse_horizontal(
                    node1, node2
                )  # always False at the moment
            if self._nv_universal_gemm_scheduling.is_nv_universal_gemm_template(node):
                return self._nv_universal_gemm_scheduling.can_fuse_horizontal(
                    node1, node2
                )  # always False at the moment
        return self._triton_scheduling.can_fuse_horizontal(node1, node2)

    def group_fn(
        self, sizes: Sequence[Sequence[_IntLike]]
    ) -> tuple[tuple[_IntLike, ...], ...]:
        return self._triton_scheduling.group_fn(sizes)

    def codegen_template(
        self,
        template_node: BaseSchedulerNode,
        epilogue_nodes: Sequence[BaseSchedulerNode],
        prologue_nodes: Sequence[BaseSchedulerNode],
    ) -> str | None:
        if self._cutlass_scheduling.is_cutlass_template(template_node):
            assert not prologue_nodes
            return self._cutlass_scheduling.codegen_template(
                template_node, epilogue_nodes, prologue_nodes
            )
        elif self._rocm_cpp_scheduling.is_rocm_cpp_template(template_node):
            assert not epilogue_nodes
            assert not prologue_nodes
            return self._rocm_cpp_scheduling.codegen_template(
                template_node, epilogue_nodes, prologue_nodes
            )
        elif self._cutedsl_scheduling.is_cutedsl_template(template_node):
            # TODO remove this when we add epilogue support
            assert not epilogue_nodes
            assert not prologue_nodes
            return self._cutedsl_scheduling.codegen_template(
                template_node, epilogue_nodes, prologue_nodes
            )
        elif self._nv_universal_gemm_scheduling.is_nv_universal_gemm_template(
            template_node
        ):
            # NVIDIA Universal GEMM doesn't support epilogue/prologue fusion yet
            assert not epilogue_nodes
            assert not prologue_nodes
            return self._nv_universal_gemm_scheduling.codegen_template(
                template_node, epilogue_nodes, prologue_nodes
            )
        else:
            return self._triton_scheduling.codegen_template(
                template_node, epilogue_nodes, prologue_nodes
            )

    def codegen_mix_order_reduction(self, node):
        return self._triton_scheduling.codegen_mix_order_reduction(node)

    def codegen_nested_reduction(self, node):
        return self._triton_scheduling.codegen_nested_reduction(node)

    def codegen_node(self, node: FusedSchedulerNode | SchedulerNode) -> None:
        return self._triton_scheduling.codegen_node(node)

    def codegen_sync(self) -> None:
        return self._triton_scheduling.codegen_sync()

    def flush(self) -> None:
        return self._triton_scheduling.flush()

    def codegen_combo_kernel(self, *args: Any, **kwargs: Any) -> None:
        return self._triton_scheduling.codegen_combo_kernel(*args, **kwargs)

    def benchmark_fused_nodes(
        self, nodes: Sequence[BaseSchedulerNode]
    ) -> tuple[float, str]:
        return self._triton_scheduling.benchmark_fused_nodes(nodes)

    def benchmark_codegened_module(self, module):
        return self._triton_scheduling.benchmark_codegened_module(module)

    def generate_kernel_code_from_nodes(
        self,
        nodes: Sequence[Any],
        benchmark_kernel: bool = False,
        hint_override: int | None = None,
    ) -> str:
        return self._triton_scheduling.generate_kernel_code_from_nodes(
            nodes, benchmark_kernel, hint_override=hint_override
        )

    def benchmark_combo_kernel(
        self, node_list: Sequence[BaseSchedulerNode], node_benchmark_results
    ) -> tuple[float, float, list[str | None]]:
        return self._triton_scheduling.benchmark_combo_kernel(
            node_list, node_benchmark_results
        )



from collections.abc import Callable
from typing import Generic, TypeVar


T = TypeVar("T")


def _value_or(opt: T | None, default: T) -> T:
    return opt if opt is not None else default


class SegmentedTree(Generic[T]):
    def __init__(
        self,
        values: list[T],
        update_op: Callable[[T, T], T],
        summary_op: Callable[[T, T], T],
        identity_element: T,
    ):
        """
        Initialize a segment tree with the given values and operations.

        Args:
            values: list of initial values
            update_op: Function to apply when updating a value (e.g., addition)
            summary_op: Function to summarize two values (e.g., min, max, sum)
            identity_element: Identity element for the summary_op (e.g., 0 for sum, float('inf') for min)

        Raises:
            ValueError: If the input values list is empty
        """
        if not values:
            raise ValueError("Cannot create a segment tree with empty values list")

        self.n = len(values)
        self.update_op = update_op
        self.summary_op = summary_op
        self.identity = identity_element

        # Size of segment tree array (next power of 2 * 2)
        # The tree follows a standard heap layout where
        # node `n`'s children are at `2*n` and `2*n+1`.
        # Index 0 is unused.
        self.size = 1
        while self.size < self.n:
            self.size *= 2
        self.size *= 2

        # Initialize tree and lazy arrays
        self.tree = [identity_element] * self.size
        # The lazy array contains updates to the given node
        # Upon update, we only push updates to the top-most
        # nodes that fully receive the update. We then
        # propagate the update down as required (i.e., when
        # we receive an interval query that neither fully
        # contains the node nor fully doesn't contain the
        # node
        self.lazy: list[T | None] = [None] * self.size

        # Build the tree
        self._build(values, 1, 0, self.n - 1)

    def _build(self, values: list[T], node: int, start: int, end: int) -> None:
        """
        Build the segment tree recursively.

        Args:
            values: Original array of values
            node: Current node index in the segment tree
            start: Start index of the segment
            end: End index of the segment
        """
        if start == end:
            # Leaf node
            if start < len(values):
                self.tree[node] = values[start]
            return

        mid = (start + end) // 2
        left_child = 2 * node
        right_child = 2 * node + 1

        # Recursively build left and right subtrees
        self._build(values, left_child, start, mid)
        self._build(values, right_child, mid + 1, end)

        # Update current node with summary of children
        self.tree[node] = self.summary_op(self.tree[left_child], self.tree[right_child])

    def _children(self, node: int) -> list[int]:
        return [2 * node, 2 * node + 1]

    def _push_lazy(self, node: int, start: int, end: int) -> None:
        """
        Push lazy updates down to children.

        Args:
            node: Current node index
            start: Start index of the segment
            end: End index of the segment
        """
        lazy_node = self.lazy[node]
        if lazy_node is None:
            return

        # Apply lazy update to current node
        self.tree[node] = self.update_op(self.tree[node], lazy_node)

        if start != end:  # Not a leaf node
            # Propagate to children
            for child in self._children(node):
                self.lazy[child] = self.update_op(
                    _value_or(self.lazy[child], self.identity), lazy_node
                )

        # Clear the lazy value
        self.lazy[node] = None

    def _update_range_helper(
        self, node: int, start: int, end: int, left: int, right: int, value: T
    ) -> None:
        """
        Helper method to update a range of values in the segment tree.

        Args:
            node: Current node index
            start: Start index of the current segment
            end: End index of the current segment
            left: Start index of the range to update
            right: End index of the range to update
            value: Value to apply to the range
        """
        # Push lazy updates before processing this node
        self._push_lazy(node, start, end)

        # No overlap
        if start > right or end < left:
            return

        # Complete overlap
        if start >= left and end <= right:
            # Apply update to current node
            self.lazy[node] = value
            self._push_lazy(node, start, end)
            return

        # Partial overlap, recurse to children
        mid = (start + end) // 2
        left_child = 2 * node
        right_child = 2 * node + 1

        self._update_range_helper(left_child, start, mid, left, right, value)
        self._update_range_helper(right_child, mid + 1, end, left, right, value)

        # Update current node based on children
        self.tree[node] = self.summary_op(self.tree[left_child], self.tree[right_child])

    def _query_range_helper(
        self, node: int, start: int, end: int, left: int, right: int
    ) -> T:
        """
        Helper method to query a range of values in the segment tree.

        Args:
            node: Current node index
            start: Start index of the current segment
            end: End index of the current segment
            left: Start index of the range to query
            right: End index of the range to query

        Returns:
            Summary value for the range
        """
        # No overlap
        if start > right or end < left:
            return self.identity

        # Push lazy updates before processing this node
        self._push_lazy(node, start, end)

        # Complete overlap
        if start >= left and end <= right:
            return self.tree[node]

        # Partial overlap, recurse to children
        mid = (start + end) // 2
        left_child = 2 * node
        right_child = 2 * node + 1

        left_result = self._query_range_helper(left_child, start, mid, left, right)
        right_result = self._query_range_helper(right_child, mid + 1, end, left, right)

        # Combine results from children
        return self.summary_op(left_result, right_result)

    def update_range(self, start: int, end: int, value: T) -> None:
        """
        Update a range of values in the segment tree.

        Args:
            start: Start index of the range to update (inclusive)
            end: End index of the range to update (inclusive)
            value: Value to apply to the range

        Raises:
            ValueError: If start > end or indices are out of bounds
        """
        if start > end:
            raise ValueError("Start index must be less than or equal to end index")

        if start < 0 or start >= self.n:
            raise ValueError(f"Start index {start} out of bounds [0, {self.n - 1}]")

        if end < 0 or end >= self.n:
            raise ValueError(f"End index {end} out of bounds [0, {self.n - 1}]")

        self._update_range_helper(1, 0, self.n - 1, start, end, value)

    def summarize_range(self, start: int, end: int) -> T:
        """
        Query a range of values in the segment tree.

        Args:
            start: Start index of the range to query (inclusive)
            end: End index of the range to query (inclusive)

        Returns:
            Summary value for the range according to the summary operation

        Raises:
            ValueError: If start > end or indices are out of bounds
        """
        if start > end:
            raise ValueError("Start index must be less than or equal to end index")

        if start < 0 or start >= self.n:
            raise ValueError(f"Start index {start} out of bounds [0, {self.n - 1}]")

        if end < 0 or end >= self.n:
            raise ValueError(f"End index {end} out of bounds [0, {self.n - 1}]")

        return self._query_range_helper(1, 0, self.n - 1, start, end)



# mypy: allow-untyped-defs
import torch

from torch._export.db.case import SupportLevel
from torch.export import Dim

class DynamicShapeRound(torch.nn.Module):
    """
    Calling round on dynamic shapes is not supported.
    """

    def forward(self, x):
        return x[: round(x.shape[0] / 2)]

x = torch.randn(3, 2)
dim0_x = Dim("dim0_x")
example_args = (x,)
tags = {"torch.dynamic-shape", "python.builtin"}
support_level = SupportLevel.NOT_SUPPORTED_YET
dynamic_shapes = {"x": {0: dim0_x}}
model = DynamicShapeRound()



# mypy: ignore-errors

import torch


# These two dicts are autogenerated with autogen/gen_dtypes.py,
# using numpy version 1.24.3.

_can_cast_dict = {
    "no": {
        torch.float16: {
            torch.float16: True,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float32: {
            torch.float16: False,
            torch.float32: True,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: True,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex128: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint8: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: True,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint16: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: True,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint32: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: True,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: True,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int8: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int16: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: True,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int32: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: True,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: True,
            torch.bool: False,
        },
        torch.bool: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: True,
        },
    },
    "equiv": {
        torch.float16: {
            torch.float16: True,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float32: {
            torch.float16: False,
            torch.float32: True,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: True,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex128: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint8: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: True,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint16: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: True,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint32: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: True,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: True,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int8: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int16: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: True,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int32: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: True,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: True,
            torch.bool: False,
        },
        torch.bool: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: False,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: True,
        },
    },
    "safe": {
        torch.float16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float32: {
            torch.float16: False,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex128: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint8: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: False,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.uint16: {
            torch.float16: False,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: False,
            torch.int16: False,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.uint32: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: True,
            torch.bool: False,
        },
        torch.uint64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: True,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.int8: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int16: {
            torch.float16: False,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int32: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: True,
            torch.complex64: False,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: True,
            torch.bool: False,
        },
        torch.bool: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
    },
    "same_kind": {
        torch.float16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float32: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.float64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex64: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.complex128: {
            torch.float16: False,
            torch.float32: False,
            torch.float64: False,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: False,
            torch.int16: False,
            torch.int32: False,
            torch.int64: False,
            torch.bool: False,
        },
        torch.uint8: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.uint16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.uint32: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.uint64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int8: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int32: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.int64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: False,
            torch.uint16: False,
            torch.uint32: False,
            torch.uint64: False,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: False,
        },
        torch.bool: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
    },
    "unsafe": {
        torch.float16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.float32: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.float64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.complex64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.complex128: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.uint8: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.uint16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.uint32: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.uint64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.int8: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.int16: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.int32: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.int64: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
        torch.bool: {
            torch.float16: True,
            torch.float32: True,
            torch.float64: True,
            torch.complex64: True,
            torch.complex128: True,
            torch.uint8: True,
            torch.uint16: True,
            torch.uint32: True,
            torch.uint64: True,
            torch.int8: True,
            torch.int16: True,
            torch.int32: True,
            torch.int64: True,
            torch.bool: True,
        },
    },
}


_result_type_dict = {
    torch.float16: {
        torch.float16: torch.float16,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.float16,
        torch.uint16: torch.float32,
        torch.uint32: torch.float64,
        torch.uint64: torch.float64,
        torch.int8: torch.float16,
        torch.int16: torch.float32,
        torch.int32: torch.float64,
        torch.int64: torch.float64,
        torch.bool: torch.float16,
    },
    torch.float32: {
        torch.float16: torch.float32,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.float32,
        torch.uint16: torch.float32,
        torch.uint32: torch.float64,
        torch.uint64: torch.float64,
        torch.int8: torch.float32,
        torch.int16: torch.float32,
        torch.int32: torch.float64,
        torch.int64: torch.float64,
        torch.bool: torch.float32,
    },
    torch.float64: {
        torch.float16: torch.float64,
        torch.float32: torch.float64,
        torch.float64: torch.float64,
        torch.complex64: torch.complex128,
        torch.complex128: torch.complex128,
        torch.uint8: torch.float64,
        torch.uint16: torch.float64,
        torch.uint32: torch.float64,
        torch.uint64: torch.float64,
        torch.int8: torch.float64,
        torch.int16: torch.float64,
        torch.int32: torch.float64,
        torch.int64: torch.float64,
        torch.bool: torch.float64,
    },
    torch.complex64: {
        torch.float16: torch.complex64,
        torch.float32: torch.complex64,
        torch.float64: torch.complex128,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.complex64,
        torch.uint16: torch.complex64,
        torch.uint32: torch.complex128,
        torch.uint64: torch.complex128,
        torch.int8: torch.complex64,
        torch.int16: torch.complex64,
        torch.int32: torch.complex128,
        torch.int64: torch.complex128,
        torch.bool: torch.complex64,
    },
    torch.complex128: {
        torch.float16: torch.complex128,
        torch.float32: torch.complex128,
        torch.float64: torch.complex128,
        torch.complex64: torch.complex128,
        torch.complex128: torch.complex128,
        torch.uint8: torch.complex128,
        torch.uint16: torch.complex128,
        torch.uint32: torch.complex128,
        torch.uint64: torch.complex128,
        torch.int8: torch.complex128,
        torch.int16: torch.complex128,
        torch.int32: torch.complex128,
        torch.int64: torch.complex128,
        torch.bool: torch.complex128,
    },
    torch.uint8: {
        torch.float16: torch.float16,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.uint8,
        torch.uint16: torch.uint16,
        torch.uint32: torch.uint32,
        torch.uint64: torch.uint64,
        torch.int8: torch.int16,
        torch.int16: torch.int16,
        torch.int32: torch.int32,
        torch.int64: torch.int64,
        torch.bool: torch.uint8,
    },
    torch.uint16: {
        torch.float16: torch.float32,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.uint16,
        torch.uint16: torch.uint16,
        torch.uint32: torch.uint32,
        torch.uint64: torch.uint64,
        torch.int8: torch.int32,
        torch.int16: torch.int32,
        torch.int32: torch.int32,
        torch.int64: torch.int64,
        torch.bool: torch.uint16,
    },
    torch.uint32: {
        torch.float16: torch.float64,
        torch.float32: torch.float64,
        torch.float64: torch.float64,
        torch.complex64: torch.complex128,
        torch.complex128: torch.complex128,
        torch.uint8: torch.uint32,
        torch.uint16: torch.uint32,
        torch.uint32: torch.uint32,
        torch.uint64: torch.uint64,
        torch.int8: torch.int64,
        torch.int16: torch.int64,
        torch.int32: torch.int64,
        torch.int64: torch.int64,
        torch.bool: torch.uint32,
    },
    torch.uint64: {
        torch.float16: torch.float64,
        torch.float32: torch.float64,
        torch.float64: torch.float64,
        torch.complex64: torch.complex128,
        torch.complex128: torch.complex128,
        torch.uint8: torch.uint64,
        torch.uint16: torch.uint64,
        torch.uint32: torch.uint64,
        torch.uint64: torch.uint64,
        torch.int8: torch.float64,
        torch.int16: torch.float64,
        torch.int32: torch.float64,
        torch.int64: torch.float64,
        torch.bool: torch.uint64,
    },
    torch.int8: {
        torch.float16: torch.float16,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.int16,
        torch.uint16: torch.int32,
        torch.uint32: torch.int64,
        torch.uint64: torch.float64,
        torch.int8: torch.int8,
        torch.int16: torch.int16,
        torch.int32: torch.int32,
        torch.int64: torch.int64,
        torch.bool: torch.int8,
    },
    torch.int16: {
        torch.float16: torch.float32,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.int16,
        torch.uint16: torch.int32,
        torch.uint32: torch.int64,
        torch.uint64: torch.float64,
        torch.int8: torch.int16,
        torch.int16: torch.int16,
        torch.int32: torch.int32,
        torch.int64: torch.int64,
        torch.bool: torch.int16,
    },
    torch.int32: {
        torch.float16: torch.float64,
        torch.float32: torch.float64,
        torch.float64: torch.float64,
        torch.complex64: torch.complex128,
        torch.complex128: torch.complex128,
        torch.uint8: torch.int32,
        torch.uint16: torch.int32,
        torch.uint32: torch.int64,
        torch.uint64: torch.float64,
        torch.int8: torch.int32,
        torch.int16: torch.int32,
        torch.int32: torch.int32,
        torch.int64: torch.int64,
        torch.bool: torch.int32,
    },
    torch.int64: {
        torch.float16: torch.float64,
        torch.float32: torch.float64,
        torch.float64: torch.float64,
        torch.complex64: torch.complex128,
        torch.complex128: torch.complex128,
        torch.uint8: torch.int64,
        torch.uint16: torch.int64,
        torch.uint32: torch.int64,
        torch.uint64: torch.float64,
        torch.int8: torch.int64,
        torch.int16: torch.int64,
        torch.int32: torch.int64,
        torch.int64: torch.int64,
        torch.bool: torch.int64,
    },
    torch.bool: {
        torch.float16: torch.float16,
        torch.float32: torch.float32,
        torch.float64: torch.float64,
        torch.complex64: torch.complex64,
        torch.complex128: torch.complex128,
        torch.uint8: torch.uint8,
        torch.uint16: torch.uint16,
        torch.uint32: torch.uint32,
        torch.uint64: torch.uint64,
        torch.int8: torch.int8,
        torch.int16: torch.int16,
        torch.int32: torch.int32,
        torch.int64: torch.int64,
        torch.bool: torch.bool,
    },
}



# mypy: allow-untyped-defs
import sys
import types
from contextlib import contextmanager

import torch


# The idea for this parameter is that we forbid bare assignment
# to torch.backends.<cudnn|mkldnn>.enabled and friends when running our
# test suite, where it's very easy to forget to undo the change
# later.
__allow_nonbracketed_mutation_flag = True


def disable_global_flags():
    global __allow_nonbracketed_mutation_flag
    __allow_nonbracketed_mutation_flag = False


def flags_frozen():
    return not __allow_nonbracketed_mutation_flag


@contextmanager
def __allow_nonbracketed_mutation():
    global __allow_nonbracketed_mutation_flag
    old = __allow_nonbracketed_mutation_flag
    __allow_nonbracketed_mutation_flag = True
    try:
        yield
    finally:
        __allow_nonbracketed_mutation_flag = old


class ContextProp:
    def __init__(self, getter, setter):
        self.getter = getter
        self.setter = setter

    def __get__(self, obj, objtype):
        return self.getter()

    def __set__(self, obj, val):
        if not flags_frozen():
            self.setter(val)
        else:
            raise RuntimeError(
                f"not allowed to set {obj.__name__} flags "
                "after disable_global_flags; please use flags() context manager instead"
            )


class PropModule(types.ModuleType):
    def __init__(self, m, name):
        super().__init__(name)
        self.m = m

    def __getattr__(self, attr):
        return self.m.__getattribute__(attr)


class _FP32Precision:
    def __init__(self, backend, op):
        self.backend = backend
        self.op = op

    def __setattr__(self, name, value):
        if name == "fp32_precision":
            torch._C._set_fp32_precision_setter(self.backend, self.op, value)
        elif name in ("backend", "op"):
            super().__setattr__(name, value)
        else:
            raise AttributeError("Unknown attribute " + name)

    def __getattr__(self, name):
        if name == "fp32_precision":
            return torch._C._get_fp32_precision_getter(self.backend, self.op)
        else:
            raise AttributeError("Unknown attribute " + name)


def set_flags(_fp32_precision="none"):
    orig_flags = (torch._C._get_fp32_precision_getter("generic", "all"),)
    if _fp32_precision is not None:
        torch._C._set_fp32_precision_setter("generic", "all", _fp32_precision)
    return orig_flags


@contextmanager
def flags(fp32_precision="none"):
    with __allow_nonbracketed_mutation():
        orig_flags = set_flags(fp32_precision)
    try:
        yield
    finally:
        with __allow_nonbracketed_mutation():
            set_flags(*orig_flags)


def _get_fp32_precision_getter(backend, op):
    def inner():
        return torch._C._get_fp32_precision_getter(backend, op)

    return inner


def _set_fp32_precision_setter(backend, op):
    def inner(precision):
        return torch._C._set_fp32_precision_setter(backend, op, precision)

    return inner


class GenericModule(PropModule):
    fp32_precision = ContextProp(
        _get_fp32_precision_getter("generic", "all"),
        _set_fp32_precision_setter("generic", "all"),
    )


sys.modules[__name__] = GenericModule(sys.modules[__name__], __name__)

from torch.backends import (
    cpu as cpu,
    cuda as cuda,
    cudnn as cudnn,
    cusparselt as cusparselt,
    kleidiai as kleidiai,
    mha as mha,
    miopen as miopen,
    mkl as mkl,
    mkldnn as mkldnn,
    mps as mps,
    nnpack as nnpack,
    openmp as openmp,
    opt_einsum as opt_einsum,
    python_native as python_native,
    quantized as quantized,
)
